-------------------------------------------------------------------------------
--  ----------------------------------------------------------------------------
-- the real lcd_clock, counters and FSM 
-------------------------------------------------------------------------------

LIBRARY ieee;
USE ieee.std_logic_1164.ALL;
USE ieee.numeric_std.ALL;

ENTITY cpld_lcd IS
  GENERIC (
    G_SIMU_ONLY   :       BOOLEAN := FALSE);
  PORT (
    clk_i         : IN    STD_LOGIC;
    rst_i         : IN    STD_LOGIC;
    --lcd interface
    lcd_data_o    : INOUT STD_LOGIC_VECTOR(7 DOWNTO 0);
    lcd_rs_o      : OUT   STD_LOGIC;
    lcd_rw_o      : OUT   STD_LOGIC;
    lcd_e_o       : OUT   STD_LOGIC;    -- 4/8 bits data -> 7/11 pins
    lcd_psb_o     : OUT   STD_LOGIC;    --parallel or serial interface
    lcd_csb_n_o   : OUT   STD_LOGIC;
    lcd_reset_n_o : OUT   STD_LOGIC
    );

END cpld_lcd;

ARCHITECTURE behave OF cpld_lcd IS
  SIGNAL sec_tick_q, min_tick_q, hrs_tick_q, day_tick_q : BOOLEAN;
  SIGNAL sec_low_co, min_low_co, hrs_low_co             : BOOLEAN;

  SIGNAL prescaler_q     : UNSIGNED(14 DOWNTO 0);
  SIGNAL all_ticks_muxed : BOOLEAN;

  SIGNAL digit_sl : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL digit_sh : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL digit_ml : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL digit_mh : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL digit_hl : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL digit_hh : STD_LOGIC_VECTOR(7 DOWNTO 0);

  SIGNAL bit_del_q                      : UNSIGNED(0 DOWNTO 0);
  SIGNAL ena_tick                       : BOOLEAN;
  SIGNAL digitcode_cmdcode_selcounter_q : UNSIGNED(2 DOWNTO 0) := "000";

  SIGNAL next_in_sequence_q, sequence_finished : BOOLEAN;
  SIGNAL start_sequence                        : BOOLEAN;

  TYPE STATE_TYPE IS (IDLE, ENA_LOW, ENA_HIGH);
  SIGNAL state_q : STATE_TYPE;

  TYPE MAIN_STATE_TYPE IS (PWRUP, INIT, UPDATE_DISPLAY, WAIT_SEC);
  SIGNAL main_state_q : MAIN_STATE_TYPE;

  SIGNAL digit_code        : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL cmd_code          : STD_LOGIC_VECTOR(7 DOWNTO 0);
  SIGNAL do_display_data_q : BOOLEAN;

  COMPONENT digit_counter_10
    PORT (
      clk_i       : IN  STD_LOGIC;
      rst_i       : IN  STD_LOGIC;
      tick_i      : IN  BOOLEAN;
      count_oq    : OUT STD_LOGIC_VECTOR(3 DOWNTO 0);
      overflow_oq : OUT BOOLEAN);
  END COMPONENT;


BEGIN  -- behave

  --lcd runs at 3v3
  --fixed signals
  lcd_reset_n_o <= '1';
  lcd_psb_o     <= '1';
  lcd_csb_n_o   <= '0';

  --all digitcounter
  sec_low : digit_counter_10
    PORT MAP (
      clk_i       => clk_i,
      rst_i       => rst_i,
      tick_i      => sec_tick_q,
      count_oq    => digit_sl(3 DOWNTO 0),
      overflow_oq => sec_low_co);

  sec_high : ENTITY work.digit_counter_6
    PORT MAP (
      clk_i       => clk_i,
      rst_i       => rst_i,
      tick_i      => sec_low_co,
      count_oq    => digit_sh(2 DOWNTO 0),
      overflow_oq => min_tick_q);

  min_low : digit_counter_10
    PORT MAP (
      clk_i       => clk_i,
      rst_i       => rst_i,
      tick_i      => min_tick_q,
      count_oq    => digit_ml(3 DOWNTO 0),
      overflow_oq => min_low_co);

  min_high : ENTITY work.digit_counter_6
    PORT MAP (
      clk_i       => clk_i,
      rst_i       => rst_i,
      tick_i      => min_low_co,
      count_oq    => digit_mh(2 DOWNTO 0),
      overflow_oq => hrs_tick_q);

  hrs_low : digit_counter_10
    PORT MAP (
      clk_i       => clk_i,
      rst_i       => rst_i,
      tick_i      => hrs_tick_q,
      count_oq    => digit_hl(3 DOWNTO 0),
      overflow_oq => hrs_low_co);

  hrs_high : ENTITY work.digit_counter_2
    PORT MAP (
      clk_i       => clk_i,
      rst_i       => rst_i,
      tick_i      => hrs_low_co,
      count_oq    => digit_hh(1 DOWNTO 0),
      overflow_oq => day_tick_q);

--fill bits TO LCD character code

  digit_sl(7 DOWNTO 4) <= "0011";
  digit_ml(7 DOWNTO 4) <= "0011";
  digit_hl(7 DOWNTO 4) <= "0011";
  digit_sh(7 DOWNTO 3) <= "00110";
  digit_mh(7 DOWNTO 3) <= "00110";
  digit_hh(7 DOWNTO 2) <= "001100";


  -- command codes
  WITH digitcode_cmdcode_selcounter_q SELECT
    --Function set "8 bit interface" "single line,display,font 5x7"
    cmd_code <= "00110001" WHEN "000",
    --bias set: BS: 1/5; display has 2 lines
    "00010100"             WHEN "001",
    --Power control: ICON disp off;Booster on; contrast
    "01010101"             WHEN "010",
    --follower (spannungsfolger) control: follower circuit on, set amplif. ratio
    "01101101"             WHEN "011",
    --set contrast: contrast set to 1000
    "01111000"             WHEN "100",
    --display on, no cursor
    "00001100"             WHEN "101",
    --clear display and cursor home
    "00000001"             WHEN "110",
    --entry mode set
    "00000110"             WHEN OTHERS;

  --display off
--    "00001000"             when "001",

  --digit codes
  WITH digitcode_cmdcode_selcounter_q SELECT
    digit_code <= digit_sl WHEN "111",
    digit_sh               WHEN "110",
    digit_ml               WHEN "100",
    digit_mh               WHEN "011",
    digit_hl               WHEN "001",
    digit_hh               WHEN "000",
    "00111010"             WHEN OTHERS;  -- code for ":"


  --digit_number selects digit to display, is counter
  PROCESS(clk_i, rst_i)
  BEGIN
    IF rst_i = '1' THEN
      digitcode_cmdcode_selcounter_q     <= "000";
    ELSIF rising_edge(clk_i) THEN
      IF next_in_sequence_q THEN
        IF digitcode_cmdcode_selcounter_q = "111" THEN
          digitcode_cmdcode_selcounter_q <= "000";
        ELSE
          digitcode_cmdcode_selcounter_q <= digitcode_cmdcode_selcounter_q + 1;
        END IF;
      END IF;
    END IF;
  END PROCESS;
  sequence_finished                      <= (digitcode_cmdcode_selcounter_q = "111" AND do_display_data_q)
                                            OR (digitcode_cmdcode_selcounter_q = "111" AND NOT do_display_data_q);


--main FSM
  PROCESS(clk_i, rst_i)
  BEGIN
    IF rst_i = '1' THEN
      main_state_q         <= PWRUP;
      do_display_data_q    <= FALSE;
      start_sequence       <= FALSE;
    ELSIF rising_edge(clk_i) THEN
      start_sequence       <= FALSE;
      CASE main_state_q IS
        WHEN PWRUP =>
          IF sec_tick_q THEN
            main_state_q   <= INIT;
            start_sequence <= TRUE;
          END IF;

        WHEN INIT =>
          IF sequence_finished THEN
            do_display_data_q <= TRUE;
            start_sequence    <= TRUE;
            main_state_q      <= UPDATE_DISPLAY;
          END IF;

        WHEN UPDATE_DISPLAY =>
          IF sequence_finished THEN
            do_display_data_q <= FALSE;
            main_state_q      <= WAIT_SEC;
          END IF;

        WHEN WAIT_SEC =>
          IF sec_tick_q THEN
            start_sequence    <= TRUE;
            do_display_data_q <= TRUE;
            main_state_q      <= UPDATE_DISPLAY;
          END IF;

        WHEN OTHERS => NULL;
      END CASE;
    END IF;
  END PROCESS;


--FSM for cmd and data seqences
  PROCESS(clk_i, rst_i)
  BEGIN
    IF rst_i = '1' THEN
      state_q            <= IDLE;
      next_in_sequence_q <= FALSE;
      lcd_e_o            <= '0';
    ELSIF rising_edge(clk_i) THEN
      next_in_sequence_q <= FALSE;
      lcd_e_o            <= '0';
      CASE state_q IS
        WHEN IDLE =>
          lcd_e_o        <= '0';
          IF start_sequence THEN
            lcd_e_o   <= '1';
            state_q      <= ENA_HIGH;
          END IF;

        WHEN ENA_HIGH =>
          lcd_e_o   <= '1';
          IF ena_tick THEN
            state_q <= ENA_LOW;
          END IF;

        WHEN ENA_LOW =>
          lcd_e_o                <= '0';
          IF ena_tick THEN
            IF sequence_finished THEN
              state_q            <= IDLE;
            ELSE
              next_in_sequence_q <= TRUE;
              state_q            <= ENA_HIGH;
            END IF;
          END IF;

        WHEN OTHERS => NULL;
      END CASE;
    END IF;
  END PROCESS;


  PROCESS(clk_i, rst_i)
  BEGIN
    IF rst_i = '1' THEN
      lcd_data_o   <= ( OTHERS => '0' );
      lcd_rs_o     <= '0';
      lcd_rw_o     <= '0';
    ELSIF rising_edge(clk_i) THEN
      lcd_rw_o     <= '0';
      IF do_display_data_q THEN
        lcd_data_o <= digit_code;
        lcd_rs_o   <= '1';
      ELSE
        lcd_data_o <= cmd_code;
        lcd_rs_o   <= '0';
      END IF;
    END IF;
  END PROCESS;

-- lcd_data_o <= digit_code when do_display_data_q else cmd_code;
-- lcd_rs_o <= '1' when do_display_data_q else '0';
-- lcd_rw_o <= '0';

--prescaler (based on 32.xyz kHz Uhrenquartz)
--period 30.5 us
  PROCESS(clk_i, rst_i)
  BEGIN
    IF rst_i = '1' THEN
      prescaler_q   <= (OTHERS => '0');
      sec_tick_q    <= FALSE;
    ELSIF rising_edge(clk_i) THEN
      sec_tick_q    <= FALSE;
      bit_del_q(0)  <= prescaler_q(0);
      IF (NOT G_SIMU_ONLY AND prescaler_q = "111111111111111") OR
        ( G_SIMU_ONLY AND prescaler_q(7 DOWNTO 0) = "11111111") THEN
        sec_tick_q  <= TRUE;
        prescaler_q <= "000000000000000";
      ELSE
        prescaler_q <= prescaler_q + 1;
      END IF;
    END IF;
  END PROCESS;
  ena_tick          <= prescaler_q(0) /= bit_del_q(0);


END behave;



